use std::sync::Arc;

use log::debug;
use mongodb::bson::{doc, to_bson};

use crate::database::{models::Guild, Client as DatabaseClient};

use serenity::{
    async_trait,
    model::prelude::{
        command::CommandOptionType,
        interaction::{
            application_command::{ApplicationCommandInteraction, CommandDataOptionValue},
            InteractionResponseType,
        },
    },
    prelude::Context,
};

use super::{
    commons::{BotCommandOption, Error},
    BotCommand,
};

pub struct DeleteTagCommand {
    interaction: ApplicationCommandInteraction,
}

#[async_trait]
impl BotCommand for DeleteTagCommand {
    fn new(interaction: ApplicationCommandInteraction) -> Self {
        debug!("Creating a new DeleteTagCommand object");
        DeleteTagCommand { interaction }
    }

    fn name() -> String {
        String::from("delete_tag")
    }

    fn description() -> String {
        String::from("Delete a tag from the server")
    }

    fn options_list() -> Vec<BotCommandOption> {
        vec![BotCommandOption {
            name: String::from("tag"),
            description: String::from("The tag to delete"),
            kind: CommandOptionType::String,
            required: true,
        }]
    }

    async fn run(&self, context: Context, database: Arc<DatabaseClient>) -> Result<(), Error> {
        let tag_name = match self.interaction.data.options.first() {
            Some(a) => match &a.resolved {
                Some(r) => match r {
                    CommandDataOptionValue::String(r_str) => Ok(r_str),
                    _ => Err(Error::ArgumentDeserialization(
                        "Received non String argument for DeleteTagCommand".to_string(),
                    )),
                },
                None => Err(Error::ArgumentDeserialization(
                    "Failed to deserialize argument for DeleteTagCommand".to_string(),
                )),
            },
            None => Err(Error::InvalidCommand(
                "Failed to find argument in DeleteTagCommand".to_string(),
            )),
        }?;

        let guild_id = match self.interaction.guild_id {
            Some(r) => Ok(r.to_string()),
            None => Err(Error::ContextRetrieval(
                "Failed to extract guild id from current context".to_string(),
            )),
        }?;

        let guild = match database.get_by_id::<Guild>(&guild_id).await {
            Ok(query) => match query {
                Some(r) => Ok(r),
                None => Err(Error::ContextRetrieval(
                    "Failed to retrieve the guild where the command was issued".to_string(),
                )),
            },
            Err(()) => Err(Error::DatabaseQuery(
                "Failed to access to the database".to_string(),
            )),
        }?;

        let response: String =
            if let Some(tag_index) = guild.tags.iter().position(|t| t.name == *tag_name) {
                let mut clone_tags = guild.tags.clone();
                clone_tags.remove(tag_index);

                match database
                    .update_one::<Guild>(
                        guild,
                        doc! {"$set": { "tags": to_bson(&clone_tags).unwrap() }},
                    )
                    .await
                {
                    Ok(_) => Ok(String::from("Successfully remove the tag")),
                    Err(()) => Err(Error::DatabaseQuery(
                        "Failed to remove tag from the database".to_string(),
                    )),
                }?
            } else {
                String::from("No matching tag for this server.")
            };

        self.interaction
            .create_interaction_response(context.http, |r| {
                r.kind(InteractionResponseType::ChannelMessageWithSource)
                    .interaction_response_data(|message| message.content(response))
            })
            .await
            .map_err(Error::DiscordAPICall)
    }
}
