use std::{collections::HashSet, sync::Arc};

use log::debug;
use mongodb::bson::{doc, to_bson};

use serenity::{
    async_trait,
    model::prelude::interaction::{
        application_command::ApplicationCommandInteraction, InteractionResponseType,
    },
    prelude::Context,
};

use super::{
    commons::{BotCommandOption, Error},
    BotCommand,
};
use crate::{
    database::{client::Client as DatabaseClient, models::Guild},
    discord::message_builders::selector_builder::EmbedSelector,
};

pub struct SubscribeCommand {
    interaction: ApplicationCommandInteraction,
}

#[async_trait]
impl BotCommand for SubscribeCommand {
    fn new(interaction: ApplicationCommandInteraction) -> Self {
        debug!("Creating a new SubscribeCommand object");
        SubscribeCommand { interaction }
    }

    fn name() -> String {
        String::from("subscribe")
    }

    fn description() -> String {
        String::from("Subscribe to a selection of tags")
    }

    fn options_list() -> Vec<BotCommandOption> {
        vec![]
    }

    async fn run(&self, context: Context, database: Arc<DatabaseClient>) -> Result<(), Error> {
        self.interaction
            .create_interaction_response(&context.http, |response| {
                response.kind(InteractionResponseType::DeferredChannelMessageWithSource)
            })
            .await
            .map_err(Error::DiscordAPICall)?;

        let user_id = self.interaction.user.id.to_string();
        let guild_id = match self.interaction.guild_id {
            Some(id) => Ok(id.to_string()),
            None => Err(Error::ContextRetrieval(
                "Failed to extract guild id from current context".to_string(),
            )),
        }?;

        let guild = match database.get_by_id::<Guild>(&guild_id).await {
            Ok(query) => match query {
                Some(r) => Ok(r),
                None => Err(Error::ContextRetrieval(
                    "Failed to retrieve the guild where the command was issued".to_string(),
                )),
            },
            Err(()) => Err(Error::DatabaseQuery(
                "Failed to access to the database".to_string(),
            )),
        }?;

        let mut available_tags_str: Vec<String> = Vec::new();
        let mut user_subscriptions: HashSet<String> = HashSet::new();

        for tag in guild.tags.as_slice() {
            available_tags_str.push(tag.name.clone());
            if tag.subscribers.contains(&user_id) {
                user_subscriptions.insert(tag.name.clone());
            }
        }

        let mut selector = EmbedSelector::new(
            "Select the tags you want to subscribe to".to_string(),
            Self::description(),
            &self.interaction,
            &context,
            available_tags_str,
            Some(user_subscriptions),
        );

        let user_selection = selector.get_user_selection().await?;

        if let Some(selection) = user_selection {
            let mut cloned_tags = guild.tags.clone();
            for t in cloned_tags.as_mut_slice() {
                if selection.contains(&t.name) && !t.subscribers.contains(&user_id) {
                    t.subscribers.push(user_id.clone());
                } else if !selection.contains(&t.name) && t.subscribers.contains(&user_id) {
                    t.subscribers.retain(|x| *x != user_id);
                }
            }

            database
                .update_one::<Guild>(
                    guild,
                    doc! {"$set": {"tags": to_bson(&cloned_tags.to_vec()).unwrap()}},
                )
                .await
                .map_err(|_| {
                    Error::DatabaseQuery(
                        "Failed to update user subscriptions in database".to_string(),
                    )
                })?;
        }

        let mut response = match self.interaction.get_interaction_response(&context).await {
            Ok(r) => Ok(r),
            Err(_e) => Err(Error::ContextRetrieval(
                "Failed to fetch initial interaction response".to_string(),
            )),
        }?;

        response
            .delete_reactions(&context)
            .await
            .map_err(Error::DiscordAPICall)?;

        response
            .edit(&context, |msg| msg.suppress_embeds(true).content("Done !"))
            .await
            .map_err(Error::DiscordAPICall)?;

        Ok(())
    }
}
