use std::{collections::HashSet, sync::Arc};

use log::debug;

use serenity::{
    async_trait,
    model::prelude::{
        interaction::{
            application_command::ApplicationCommandInteraction, InteractionResponseType,
        },
        UserId,
    },
    prelude::{Context, Mentionable},
};

use super::{
    commons::{BotCommandOption, Error},
    BotCommand,
};

use crate::{
    database::{models::Guild, Client as DatabaseClient},
    discord::message_builders::selector_builder::EmbedSelector,
};

pub struct TagNotifyCommand {
    interaction: ApplicationCommandInteraction,
}

#[async_trait]
impl BotCommand for TagNotifyCommand {
    fn name() -> String {
        debug!("Creating a new TagNotifyCommand object");
        String::from("notify")
    }

    fn new(interaction: ApplicationCommandInteraction) -> Self {
        TagNotifyCommand { interaction }
    }

    fn description() -> String {
        String::from("Ping users according to a list of tag")
    }

    fn options_list() -> Vec<BotCommandOption> {
        vec![]
    }

    async fn run(&self, context: Context, database: Arc<DatabaseClient>) -> Result<(), Error> {
        self.interaction
            .create_interaction_response(&context.http, |response| {
                response.kind(InteractionResponseType::DeferredChannelMessageWithSource)
            })
            .await
            .map_err(Error::DiscordAPICall)?;

        let guild_id = match self.interaction.guild_id {
            Some(id) => Ok(id.to_string()),
            None => Err(Error::ContextRetrieval(
                "Failed to extract guild id from current context".to_string(),
            )),
        }?;

        let guild = match database.get_by_id::<Guild>(&guild_id).await {
            Ok(query) => match query {
                Some(r) => Ok(r),
                None => Err(Error::ContextRetrieval(
                    "Failed to retrieve the guild where the command was issued".to_string(),
                )),
            },
            Err(()) => Err(Error::DatabaseQuery(
                "Failed to access to the database".to_string(),
            )),
        }?;

        let mut available_tags_str: Vec<String> = Vec::new();

        for tag in guild.tags.as_slice() {
            available_tags_str.push(tag.name.clone());
        }

        let mut selector = EmbedSelector::new(
            "Select the tags to notify".to_string(),
            Self::description(),
            &self.interaction,
            &context,
            available_tags_str,
            None,
        );

        let user_selection = selector.get_user_selection().await?;

        if let Some(selection) = user_selection {
            let mut users_to_ping: HashSet<String> = HashSet::new();
            let mut answer = String::new();
            for selected_tag in selection {
                let t = match guild.tags.iter().find(|s| s.name == selected_tag) {
                    Some(t) => Ok(t),
                    None => Err(Error::InvalidCommand(
                        "No matching tag found for selection".to_string(),
                    )),
                }?;

                for user_id in t.subscribers.as_slice() {
                    match user_id.parse::<u64>() {
                        Ok(id) => match &UserId(id).to_user(&context).await {
                            Ok(e) => {
                                users_to_ping.insert(e.mention().to_string());
                            }
                            Err(_e) => {}
                        },
                        Err(_e) => {}
                    }
                }
            }

            for user in users_to_ping {
                answer.push_str(user.as_str());
            }

            let response = match self.interaction.get_interaction_response(&context).await {
                Ok(r) => Ok(r),
                Err(_e) => Err(Error::ContextRetrieval(
                    "Failed to fetch initial interaction response".to_string(),
                )),
            }?;

            response
                .delete(&context)
                .await
                .map_err(Error::DiscordAPICall)?;

            // We have to create a new message as editing the original response will not notify
            // the pinged users
            self.interaction
                .create_followup_message(&context, |msg| {
                    msg.content(if answer.is_empty() {
                        "Nobody to ping for your selection".to_string()
                    } else {
                        answer
                    })
                })
                .await
                .map_err(Error::DiscordAPICall)?;
        } else {
            self.interaction
                .delete_original_interaction_response(&context)
                .await
                .map_err(Error::DiscordAPICall)?;
        }

        Ok(())
    }
}
