use std::collections::{HashMap, HashSet};

use serenity::{
    builder::{CreateEmbed, CreateEmbedAuthor, CreateEmbedFooter},
    prelude::Context,
};

use crate::discord::commands::commons::Error;

const HTML_COLOR_CODE: u32 = 0xffffff;

pub struct EmbedMessageBuilder {
    color_code: u32,
    embed_author: String,
    embed_avatar_url: String,
}

impl EmbedMessageBuilder {
    pub async fn new(context: &Context) -> Result<Self, Error> {
        let bot_user =
            context.http.get_current_user().await.map_err(|_e| {
                Error::ContextRetrieval("Failed to get current bot user".to_string())
            })?;

        let embed_author = bot_user.name.clone();

        let embed_avatar_url = bot_user
            .avatar_url()
            .unwrap_or_else(|| bot_user.default_avatar_url());

        Ok(EmbedMessageBuilder {
            color_code: HTML_COLOR_CODE,
            embed_author,
            embed_avatar_url,
        })
    }

    fn create_embed_author(&self) -> CreateEmbedAuthor {
        CreateEmbedAuthor(HashMap::new())
            .name(self.embed_author.clone())
            .icon_url(self.embed_avatar_url.clone())
            .to_owned()
    }

    fn create_embed_base(&self) -> CreateEmbed {
        CreateEmbed(HashMap::new())
            .set_author(self.create_embed_author())
            .colour(self.color_code)
            .to_owned()
    }

    fn create_embed_pages_footer(
        &self,
        current_page: usize,
        total_pages: usize,
    ) -> CreateEmbedFooter {
        CreateEmbedFooter(HashMap::new())
            .text(format!("Page {current_page}/{total_pages}"))
            .to_owned()
    }

    pub fn create_bot_credentials_embed(&self) -> CreateEmbed {
        self.create_embed_base()
            .title("Yorokobot")
            .fields(vec![
                ("Version", env!("CARGO_PKG_VERSION"), false),
                (
                    "License",
                    "The source code is under the European Union Public License v1.2",
                    false,
                ),
                ("Source code", env!("CARGO_PKG_REPOSITORY"), false),
                (
                    "Illustrator's Twitter",
                    "https://twitter.com/MaewenMitzuki",
                    false,
                ),
            ])
            .to_owned()
    }

    pub fn create_selection_embed(
        &self,
        title: &str,
        description: &str,
        selectable: &[String],
        selected: &HashSet<String>,
        pages: usize,
        current_page: usize,
    ) -> CreateEmbed {
        let mut content = "".to_string();

        const SELECTION_EMOTES: [&str; 10] = [
            ":zero:", ":one:", ":two:", ":three:", ":four:", ":five:", ":six:", ":seven:",
            ":eight:", ":nine:",
        ];

        for (value, emote) in selectable.iter().zip(SELECTION_EMOTES.iter()) {
            content.push_str(
                format!(
                    "{emote} - *{value}* {}\n",
                    if selected.contains(value) {
                        ":white_check_mark:"
                    } else {
                        ""
                    }
                )
                .as_str(),
            );
        }

        self.create_embed_base()
            .title(title)
            .description(description)
            .field("Selection", content, false)
            .set_footer(self.create_embed_pages_footer(current_page, pages))
            .to_owned()
    }
}
